#include <iostream>
#include <vector>
#include <fstream>
#include <cstdlib> // for atoi

#include "finite_field.h"
#include "matrix.h"



extern "C" {
#include "cliquer.h"
#include "cliquer_mods.h"
}

using namespace std;



FiniteField F2(2);
FiniteField F3(3);
FiniteField F5(5);
FiniteField F7(7);

const int Poly_4[]={1,1};
FiniteField F4(2,2,Poly_4);

const int Poly_9[]={1,0};
FiniteField F9(3,2,Poly_9);

const int Poly_16[]={1,1,1,1};
FiniteField F16(2,4,Poly_16);

const int Poly_25[]={3,0};
FiniteField F25(5,2,Poly_25);

const int Poly_49[]={1,0};
FiniteField F49(7,2,Poly_49);

FiniteField F(F25);

void clique_find_all_including_start_set(graph_t* const graph, int number, int* const start_set, clique_options *my_opts);
int matrix_index(Matrix<FieldElement<&F> > M, vector<Matrix<FieldElement<&F> > > matrices);


int main(int argc, char *argv[])
{
    if (argc != 6)
    {
        cout << "Usage: "<< argv[0] << " linenumber startset[4]" << endl;
        return 1;
    }

    vector<FieldElement<&F> > field(F.number_of_field_elements);
    for (int k=1; k<F.number_of_field_elements; ++k)
    {
        field[k] = field[k-1];

        ++field[k][0];

        for (int d=0; d< (const signed int) F.deg-1; ++d)
        {
            if (field[k][d] == (const signed int) F.p)
            {
                field[k][d] = 0;
                ++field[k][d+1];
            }
        }
    }


    cout << "Körper angelegt" << endl;


/** Symmetric Matrices */
/*
    unsigned int matrix_size=3;
    unsigned int distance=2;

    cout << "F" << F.number_of_field_elements << ", " << matrix_size << "x" << matrix_size << " (symmetric)" << endl << endl;

    Matrix<FieldElement<&F> > M(matrix_size, matrix_size);

    int number_of_symm_matrices = pow(F.number_of_field_elements,(matrix_size*(matrix_size+1))/2);
    vector<Matrix<FieldElement<&F> > > Mat_symm(number_of_symm_matrices, M);

    {
        vector<int> M_index((matrix_size*(matrix_size+1))/2, 0);
        vector<vector<int> > Mat_index(number_of_symm_matrices, M_index);

        for (int k=1; k<number_of_symm_matrices; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< (const signed int) (matrix_size*(matrix_size+1))/2-1; ++d)
            {
                if (Mat_index[k][d] == (const signed int) F.number_of_field_elements)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }

        for (int k=1; k<number_of_symm_matrices; ++k)
        {
            int j=0;
            for (unsigned int r=0; r<matrix_size; ++r)
            {
                Mat_symm[k][r][r]= field[Mat_index[k][j]];
                ++j;
                for (unsigned int c=0; c<r; ++c)
                {
                    Mat_symm[k][r][c]= field[Mat_index[k][j]];
                    Mat_symm[k][c][r]= field[Mat_index[k][j]];
                    ++j;
                }
            }
        }
    }

    cout << "symmetrische Matrizen angelegt" << endl;


    graph_t *graph_symm_matrices;
    graph_symm_matrices = graph_new(number_of_symm_matrices);

    cout << "Graph angelegt; Graph enthält "<< number_of_symm_matrices << " Ecken" << endl;


    for (int k=1; k<number_of_symm_matrices; ++k)
    {
        for (int l=0; l<k; ++l)
        {
            if ((Mat_symm[k]-Mat_symm[l]).rank_geq(distance))
                GRAPH_ADD_EDGE(graph_symm_matrices,k,l);
        }
    }

    cout << "Graph verknüpft; Graph enthält "<< graph_edge_count(graph_symm_matrices) << " Kanten" << endl;
*/


/** Hermitian Matrices */

    unsigned int matrix_size=2;
    unsigned int distance=2;

    cout << "F" << F.number_of_field_elements << ", " << matrix_size << "x" << matrix_size << " (hermitian)" << endl << endl;

    int q = sqrt(F.number_of_field_elements); // F = F_{q^2}

    vector<FieldElement<&F> > field_conjugated(F.number_of_field_elements);
    vector<FieldElement<&F> > f_q;
    for (int k=0; k<F.number_of_field_elements; ++k)
    {
        field_conjugated[k] = pow(field[k],q);
        if (field[k]==field_conjugated[k])
            {f_q.push_back(field[k]); cout << field[k] << endl;}
    }

    Matrix<FieldElement<&F> > M(matrix_size, matrix_size);

    int number_of_triangles = pow(F.number_of_field_elements,(matrix_size*(matrix_size-1))/2);
    int number_of_diagonals = pow(q,matrix_size);
    int number_of_herm_matrices = number_of_triangles * number_of_diagonals;
    vector<Matrix<FieldElement<&F> > > Mat_herm(number_of_herm_matrices, M);

    {
        vector<int> M_index((matrix_size*(matrix_size-1))/2, 0);
        vector<vector<int> > Mat_index(number_of_triangles, M_index);

        for (int k=1; k<number_of_triangles; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< (const signed int) (matrix_size*(matrix_size-1))/2-1; ++d)
            {
                if (Mat_index[k][d] == (const signed int) F.number_of_field_elements)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }

        for (int k=1; k<number_of_triangles; ++k)
        {
            int j=0;
            for (unsigned int r=0; r<matrix_size; ++r)
            {
               // Mat_herm[k][r][r]= 0; entry is already zero
                for (unsigned int c=0; c<r; ++c)
                {
                    Mat_herm[k][r][c]= field[Mat_index[k][j]];
                    Mat_herm[k][c][r]= field_conjugated[Mat_index[k][j]];
                    ++j;
                }
            }
        }
    }
    {
        vector<int> M_index(matrix_size, 0);
        vector<vector<int> > Mat_index(number_of_diagonals, M_index);

        for (int k=1; k<number_of_diagonals; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< matrix_size-1; ++d)
            {
                if (Mat_index[k][d] == q)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }


        for (int k=1; k<number_of_diagonals; ++k)
        {
            for (int j=0; j<number_of_triangles; ++j)
            {
                Mat_herm[j+k*number_of_triangles] = Mat_herm[j];
                for (int i=0; i<matrix_size; ++i)
                    Mat_herm[j+k*number_of_triangles][i][i] = f_q[Mat_index[k][i]];
            }
        }
    }

    cout << "hermitesche Matrizen angelegt" << endl;


    graph_t *graph_herm_matrices;
    graph_herm_matrices = graph_new(number_of_herm_matrices);


    cout << "Graph angelegt; Graph enthält "<< number_of_herm_matrices << " Ecken" << endl;

    for (int k=1; k<number_of_herm_matrices; ++k)
    {
        for (int l=0; l<k; ++l)
        {
            if ((Mat_herm[k]-Mat_herm[l]).rank_geq(distance))
                GRAPH_ADD_EDGE(graph_herm_matrices,k,l);
        }
    }
    cout << "Graph verknüpft; Graph enthält "<< graph_edge_count(graph_herm_matrices) << " Kanten" << endl;


    static clique_options clique_my_options_struct = {
	reorder_by_greedy_coloring, NULL, clique_print_time_modified, NULL, save_clique, NULL, NULL, 0
    };
    clique_options *my_opts=&clique_my_options_struct;



    CLIQUE_OUTPUTFILE.open(("cliquesF25_z"+ (string)argv[1] +".txt").c_str());

    int start_set[4];

    start_set[0] = atoi(argv[2]);
    start_set[1] = atoi(argv[3]);
    start_set[2] = atoi(argv[4]);
    start_set[3] = atoi(argv[5]);

    time_t timer1 = time(NULL);

    clique_find_all_including_start_set(graph_herm_matrices, 4, start_set, my_opts);


    cout << difftime(time(NULL),timer1) << " seconds for clique search" << endl;
    CLIQUE_OUTPUTFILE.close();
    graph_free(graph_herm_matrices);


    return 0;
}

void clique_find_all_including_start_set(graph_t* const graph, int number, int* const start_set, clique_options *my_opts)
{
    int found_size = number;

    vector<int> vertices_adjacent_to_start_set;
    for (int i=0; i<graph->n; ++i)
    {
        bool adjacent = true;
        for (int j=0; j<number; ++j)
        {
            adjacent = adjacent && GRAPH_IS_EDGE_FAST(graph,i,start_set[j]);
        }
        if (adjacent)
            vertices_adjacent_to_start_set.push_back(i);
    }

    int number_of_adjacent_vertices = vertices_adjacent_to_start_set.size();
    if (number_of_adjacent_vertices!=0)
    {
        graph_t *graph_of_adjacent_vertices;
        graph_of_adjacent_vertices = graph_new(number_of_adjacent_vertices);

        for (int i=1; i<number_of_adjacent_vertices; ++i)
        {
            for (int j=0; j<i; ++j)
            {
                if (GRAPH_IS_EDGE_FAST(graph,vertices_adjacent_to_start_set[i],vertices_adjacent_to_start_set[j]))
                    GRAPH_ADD_EDGE(graph_of_adjacent_vertices,i,j);
            }
        }


        int* userdata = new int[1+number+number_of_adjacent_vertices];
        userdata[0] = number;
        for (int k=1;k<=number;++k)
            userdata[k]=start_set[k-1];
        for (int l=0;l<number_of_adjacent_vertices;++l)
            userdata[number+1+l]=vertices_adjacent_to_start_set[l];

        my_opts->user_data = userdata;

        clique_find_all(graph_of_adjacent_vertices,0,0,true,my_opts);



        graph_free(graph_of_adjacent_vertices);
        delete[] userdata;

    }

    return;
}



int matrix_index(Matrix<FieldElement<&F> > M, vector<Matrix<FieldElement<&F> > > matrices)
{
    for (int i=0; i<matrices.size(); ++i)
    {
        if(M==matrices[i])
            return i;
    }
    return -1;
}


