#include <iostream>
#include <cstdlib>
#include <ctime>
#include <vector>
#include <fstream>


extern "C"{
#include "graph.h" // from Cliquer
}





using namespace std;

int main()
{
    int number = 1; // number of start vertices
    int start_set[] = {0};

    cout << "start set: ";
    for (int i=0; i<number; ++i) cout << start_set[i] << " ";
    cout << endl;


    ifstream fin("adjacent.txt");
    vector<int> vertices_adjacent_to_start_set;
    int v_in;
    while (fin >> v_in)
        vertices_adjacent_to_start_set.push_back(v_in);
    fin.close();

    // choose graph

    graph_t *g = graph_read_dimacs_file("graph.dimacsbin");
    int n = g->n;
    ASSERT(graph_test(g,NULL));

    ASSERT(n==vertices_adjacent_to_start_set.size() && n>0)


    time_t time_heur0 = time(NULL);

    // choose parameters

    int tcs = 0; // minimum cardinality to output


    // variables for the vertex set C
    int cardinality;
    int vertex_list_array[n];
    int vertex_index_array[n];

    // variable for the vertex set C'
    int intersection_cardinality;

    // variable for selectMinPenalty
    bool available[n];
    for(int i=0; i<n; ++i)
        available[i] = true;


    // initialize vertex set C to the set containing one random vertex
    for (int i=0; i<n; ++i)
        vertex_index_array[i] = -1;
    srand(time(NULL));
    vertex_list_array[0] = rand() % n;
    vertex_index_array[vertex_list_array[0]] = 0;
    cardinality = 1;

    // most recently added vertex
    int v = vertex_list_array[0];

    // variables for sets NI and NL
    int NI_cardinality;
    int NI_vertex_list_array[n];
    int NI_vertex_index_array[n];
    int NL_cardinality;
    int NL_vertex_list_array[n];
    int NL_vertex_index_array[n];
    int NL_availables[n];

    int new_NI_size;
    int poss_next_NI_size[n];
    int best_choices[n]; // array for indices such that choosing a vertex NI_vertex_list_array[best_choices[i]] would lead to a new set NI is as large as possible
    int best_choices_number = 0; // actual length of best_choices

    // initialize set NL with the whole vertex set excluding vertex v
    for (int i=0; i<n; ++i)
        NL_vertex_index_array[i] = NL_vertex_list_array[i] = i;
    NL_vertex_list_array[NL_vertex_index_array[v]] = NL_vertex_list_array[n-1];
    NL_vertex_index_array[NL_vertex_list_array[n-1]] = NL_vertex_index_array[v];
    NL_vertex_index_array[v] = -1;
    NL_cardinality = n-1;


    // initialize set NI
    NI_cardinality = 0;
    for (int i=0; i<n; ++i)
    {
        if(GRAPH_IS_EDGE_FAST(g,v,i))
        {
            NI_vertex_list_array[NI_cardinality] = i;
            NI_vertex_index_array[i] = NI_cardinality;
            ++NI_cardinality;
        }
        else
            NI_vertex_index_array[i] = -1;
    }



    while(true) // abort program when desired
    {
        expand: // expand
        while (NI_cardinality > 0)
        {
            // select randomly a vertex in NI, such that the new set NI is as large as possible
            new_NI_size = 0;

            for (int i=0; i<NI_cardinality; ++i)
            {
               // NI_vertex_list_array[i]
                poss_next_NI_size[i] = 0;
                for (int j=0; j<i; ++j)
                {
                    if(GRAPH_IS_EDGE_FAST(g,NI_vertex_list_array[i],NI_vertex_index_array[j]))
                    {
                        ++poss_next_NI_size[i];
                        ++poss_next_NI_size[j];
                    }
                }
            }
            for (int i=0; i<NI_cardinality; ++i)
            {
                if (poss_next_NI_size[i] > new_NI_size)
                {
                    new_NI_size = poss_next_NI_size[i];
                }
                if (new_NI_size==NI_cardinality-1) break;
            }
            best_choices_number = 0;
            for (int i=0; i<NI_cardinality; ++i)
            {
                if(poss_next_NI_size[i] == new_NI_size)
                {
                    best_choices[best_choices_number] = i;
                    ++best_choices_number;
                }
            }
            v = NI_vertex_list_array[best_choices[rand() % best_choices_number]];

            // make v unavailable
            available[v] = false;

            // add v to C
            vertex_list_array[cardinality] = v;
            vertex_index_array[v] = cardinality;
            ++cardinality;

            // update NL: the new set NL consists of the vertices of the old set NL that are adjacent to the new vertex and the vertices of the old set NI that are not adjacent to the new vertex
            //              first step: remove vertices not adjacent to the new vertex v
            for(int i=NL_cardinality-1; i>=0; --i) // go backwards because NL_cardinality is decreasing during the loop!
            {
                if (!GRAPH_IS_EDGE_FAST(g,v,NL_vertex_list_array[i]))
                {
                    NL_vertex_list_array[i] = NL_vertex_list_array[NL_cardinality-1];
                    NL_vertex_index_array[NL_vertex_list_array[NL_cardinality-1]] = i;
                    NL_vertex_index_array[NL_vertex_list_array[i]] = -1;
                    --NL_cardinality;
                }
            }
            //              second step: add the vertices of NI that are not adjacent to the new vertex v
            for (int i=0; i<NI_cardinality; ++i)
            {
                if (!GRAPH_IS_EDGE_FAST(g,v,NI_vertex_list_array[i]))
                {
                    NL_vertex_list_array[NL_cardinality] = NI_vertex_list_array[i];
                    NL_vertex_index_array[NI_vertex_list_array[i]] = NL_cardinality;
                    ++NL_cardinality;
                }
            }

            // update NI: remove vertices not adjacent to v
            for(int i=NI_cardinality-1; i>=0; --i) // go backwards because NI_cardinality is decreasing during the loop!
            {
                if (!GRAPH_IS_EDGE_FAST(g,v,NI_vertex_list_array[i]))
                {
                    NI_vertex_list_array[i] = NI_vertex_list_array[NI_cardinality-1];
                    NI_vertex_index_array[NI_vertex_list_array[NI_cardinality-1]] = i;
                    NI_vertex_index_array[NI_vertex_list_array[i]] = -1;
                    --NI_cardinality;
                }
            }
        }

        // check if current clique size is of interest and increase target clique size
        if (cardinality >= tcs)
        {
            cout << "Clique of size " << cardinality + number << " found (" << time(NULL)-time_heur0 << " s ):" << endl;
            for (int i=0; i<number; ++i)      cout << start_set[i] << " ";
            for (int i=0; i<cardinality; ++i) cout << vertices_adjacent_to_start_set[vertex_list_array[i]] << " ";
            cout << endl << endl;
            tcs = cardinality + 1;
        }



        // plateau search

        // C' = C
        intersection_cardinality = cardinality;

        while(NL_cardinality>0 && intersection_cardinality>0)
        {
            // select randomly an available vertex in NL
            int available_number = 0;

            for (int i=0; i<NL_cardinality; ++i)
            {
                if(available[NL_vertex_list_array[i]])
                {
                    NL_availables[available_number] = i;
                    ++available_number;
                }
            }
            if (available_number == 0) goto restart;
            v = NL_vertex_list_array[NL_availables[rand() % available_number]];

            // make v unavailable
            available[v] = false;


            // add v to C and remove the vertex not adjacent to v
            int w; // vertex to be exchanged by v
            for (int i=0; i<cardinality; ++i)
            {
                if(!GRAPH_IS_EDGE_FAST(g,v,vertex_list_array[i]))
                {
                    w = vertex_list_array[i];
                    vertex_list_array[i] = v;
                    vertex_index_array[v] = i;
                    vertex_index_array[w] = -1;
                    break;
                }
            }
            --intersection_cardinality;

            // update NI: the new set NI consists of the vertices of the old set NI that are adjacent to the new vertex and the
            //              vertices of the old set NL that are adjacent to the new vertex and not adjacent to the old vertex
            //              first step: remove vertices not adjacent to the new vertex v
            for(int i=NI_cardinality-1; i>=0; --i) // go backwards because NI_cardinality is decreasing during the loop!
            {
                if (!GRAPH_IS_EDGE_FAST(g,v,NI_vertex_list_array[i]))
                {
                    NI_vertex_list_array[i] = NI_vertex_list_array[NI_cardinality-1];
                    NI_vertex_index_array[NI_vertex_list_array[NI_cardinality-1]] = i;
                    NI_vertex_index_array[NI_vertex_list_array[i]] = -1;
                    --NI_cardinality;
                }
            }
            //              second step: add the vertices of NL that are adjacent to the new vertex v and not adjacent to the old vertex w
            for (int i=0; i<NL_cardinality; ++i)
            {
                if (GRAPH_IS_EDGE_FAST(g,v,NL_vertex_list_array[i]) && !GRAPH_IS_EDGE_FAST(g,w,NL_vertex_list_array[i]))
                {
                    NI_vertex_list_array[NI_cardinality] = NL_vertex_list_array[i];
                    NI_vertex_index_array[NL_vertex_list_array[i]] = NI_cardinality;
                    ++NI_cardinality;
                }
            }

            // update NL:
            //              first step: remove vertices that are adjacent to exactly one of the vertices v and w
            for(int i=NL_cardinality-1; i>=0; --i) // go backwards because NL_cardinality is decreasing during the loop!
            {
                bool edge_v = GRAPH_IS_EDGE_FAST(g,v,NL_vertex_list_array[i]);
                bool edge_w = GRAPH_IS_EDGE_FAST(g,w,NL_vertex_list_array[i]);
                bool to_remove = edge_v ^ edge_w;
                if (to_remove)
                {
                    NL_vertex_list_array[i] = NL_vertex_list_array[NL_cardinality-1];
                    NL_vertex_index_array[NL_vertex_list_array[NL_cardinality-1]] = i;
                    NL_vertex_index_array[NL_vertex_list_array[i]] = -1;
                    --NL_cardinality;
                }
            }
            //              second step: add vertices that are not in NL nor in NI nor in C and are adjacent to all vertices in C but one
            for(int i=0; i<n; ++i)
            {
                if(NI_vertex_index_array[i]==-1 && NL_vertex_index_array[i]==-1 && vertex_index_array[i]==-1)
                {
                    int number_of_not_adjacent_vertices_in_C = 0;
                    for(int j=0; j<cardinality; ++j)
                    {
                        if (!GRAPH_IS_EDGE_FAST(g,i,vertex_list_array[j]))
                        {
                            ++number_of_not_adjacent_vertices_in_C;
                            if (number_of_not_adjacent_vertices_in_C > 1) break;
                        }
                    }
                    if (number_of_not_adjacent_vertices_in_C == 1) // then add i to NL
                    {
                        NL_vertex_list_array[NL_cardinality] = i;
                        NL_vertex_index_array[i] = NL_cardinality;
                        ++NL_cardinality;
                    }
                }
            }

            if(NI_cardinality>0) goto expand;
        }


        restart: // perturbate


        // select randomly a vertex not in C
        int complement[n]; // array for the complement of C
        int complement_cardinality = 0; // actual length of complement
        for (int i=0; i<n; ++i)
        {
            if(vertex_index_array[i]==-1)
            {
                complement[complement_cardinality] = i;
                ++complement_cardinality;
            }
        }
        v = complement[rand() % complement_cardinality];

        // add v to C
        vertex_list_array[cardinality] = v;
        vertex_index_array[v] = cardinality;
        ++cardinality;

        // remove the vertices not adjacent to v
        for(int i=cardinality-1; i>=0; --i) // go backwards because cardinality is decreasing during the loop!
        {
            if(!GRAPH_IS_EDGE_FAST(g,v,vertex_list_array[i]))
            {
                vertex_index_array[vertex_list_array[cardinality-1]] = i;
                vertex_index_array[vertex_list_array[i]] = -1;
                vertex_list_array[i] = vertex_list_array[cardinality-1];
                --cardinality;
            }
        }

        //cout << endl << "restart, cardinality = " << cardinality << endl << endl;

        // update NI and NL:
        //              	step one: set them to the empty set
        NI_cardinality = NL_cardinality = 0;
        for (int i=0; i<n; ++i)
            NI_vertex_index_array[i] = NL_vertex_index_array[i] = -1;
        //               	step two: add vertices
        for(int i=0; i<n; ++i)
        {
            if(vertex_index_array[i]==-1)
            {
                int number_of_not_adjacent_vertices_in_C = 0;
                for(int j=0; j<cardinality; ++j)
                {
                    if (!GRAPH_IS_EDGE_FAST(g,i,vertex_list_array[j]))
                    {
                        ++number_of_not_adjacent_vertices_in_C;
                        if (number_of_not_adjacent_vertices_in_C > 1) break;
                    }
                }
                if (number_of_not_adjacent_vertices_in_C == 1) // then add i to NL
                {
                    NL_vertex_list_array[NL_cardinality] = i;
                    NL_vertex_index_array[i] = NL_cardinality;
                    ++NL_cardinality;
                }
                if (number_of_not_adjacent_vertices_in_C == 0) // then add i to NI
                {
                    NI_vertex_list_array[NI_cardinality] = i;
                    NI_vertex_index_array[i] = NI_cardinality;
                    ++NI_cardinality;
                }
            }
        }


        // make all vertices available again
        for(int i=0; i<n; ++i)
            available[i] = true;


   }
}
