#include <vector>
#include <fstream>
#include <cstdlib> // for atoi

#include "finite_field.h"
#include "matrix.h"
#include "orderly.h"


using namespace std;



FiniteField F2(2);

FiniteField F(F2);

int main(int argc, char *argv[])
{
    if (argc < 3)
    {
        cout << "Usage: "<< argv[0] << " filename startsetsize0" << endl;
      //  return 1;
    }
    int startsetsize0 = atoi(argv[2]);
    int startsetsize = startsetsize0 + 1;
    string outfilename = (string)argv[1] + "_out_";
    string readfilename = argv[1];

    time_t timer0 = time(NULL);



    vector<FieldElement<&F> > field(F.number_of_field_elements);
    for (int k=1; k<F.number_of_field_elements; ++k)
    {
        field[k] = field[k-1];

        ++field[k][0];

        for (int d=0; d< (const signed int) F.deg-1; ++d)
        {
            if (field[k][d] == (const signed int) F.p)
            {
                field[k][d] = 0;
                ++field[k][d+1];
            }
        }
    }


    cout << "Körper angelegt" << endl;


/** Symmetric Matrices */

    unsigned int matrix_size=3;
    unsigned int distance=2;

    cout << "F" << F.number_of_field_elements << ", " << matrix_size << "x" << matrix_size << " (symmetric)" << endl << endl;

    Matrix<FieldElement<&F> > M(matrix_size, matrix_size);

    int number_of_symm_matrices = pow(F.number_of_field_elements,(matrix_size*(matrix_size+1))/2);
    vector<Matrix<FieldElement<&F> > > Mat_symm(number_of_symm_matrices, M);

    {
        vector<int> M_index((matrix_size*(matrix_size+1))/2, 0);
        vector<vector<int> > Mat_index(number_of_symm_matrices, M_index);

        for (int k=1; k<number_of_symm_matrices; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< (const signed int) (matrix_size*(matrix_size+1))/2-1; ++d)
            {
                if (Mat_index[k][d] == (const signed int) F.number_of_field_elements)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }

        for (int k=1; k<number_of_symm_matrices; ++k)
        {
            int j=0;
            for (unsigned int r=0; r<matrix_size; ++r)
            {
                Mat_symm[k][r][r]= field[Mat_index[k][j]];
                ++j;
                for (unsigned int c=0; c<r; ++c)
                {
                    Mat_symm[k][r][c]= field[Mat_index[k][j]];
                    Mat_symm[k][c][r]= field[Mat_index[k][j]];
                    ++j;
                }
            }
        }
    }

    cout << "symmetrische Matrizen angelegt" << endl;


    graph_t *graph_symm_matrices;
    graph_symm_matrices = graph_new(number_of_symm_matrices);

	graph* g;
    int g_sz=0;
    int n = number_of_symm_matrices;
    int m =SETWORDSNEEDED(2*n);
    DYNALLOC2(graph,g,g_sz,m,2*n,"malloc");

    EMPTYGRAPH(g,m,2*n);

    cout << "Graph angelegt; Graph enthält "<< number_of_symm_matrices << " Ecken" << endl;


    for (int k=0; k<number_of_symm_matrices; ++k)
    {
        for (int l=0; l<k; ++l)
        {
            if ((Mat_symm[k]-Mat_symm[l]).rank_geq(distance))
                GRAPH_ADD_EDGE(graph_symm_matrices,k,l);
			if (!((Mat_symm[k]-Mat_symm[l]).rank_geq(2)))
                {ADDONEEDGE(g,k,l,m);}  //rd 1
            else if (!((Mat_symm[k]-Mat_symm[l]).rank_geq(3)))
                ADDONEEDGE(g,k+number_of_symm_matrices,l+number_of_symm_matrices,m); //rd 2
        }
        ADDONEEDGE(g,k,k+number_of_symm_matrices,m); // connect the two layers
    }


    cout << "Graph verknüpft; Graph enthält "<< graph_edge_count(graph_symm_matrices) << " Kanten" << endl;


    vector <Vertexset> Sk;

    ifstream fin((readfilename).c_str());
    int in;
    int counter = 0;

    ofstream fout1((outfilename + "1").c_str());
    ofstream fout2((outfilename + "2").c_str());
    ofstream fout3((outfilename + "3").c_str());
    ofstream fout4((outfilename + "4").c_str());
    ofstream fout5((outfilename + "5").c_str());
    ofstream fout6((outfilename + "6").c_str());
    ofstream fout7((outfilename + "7").c_str());
    ofstream fout8((outfilename + "8").c_str());

    Vertexset emptyset(g, graph_symm_matrices);
    Vertexset X(emptyset);

    int anz=0;

    if (!startsetsize0)
    {
        Sk = augment_vertexset(X);

        for (int j=0; j<Sk.size(); ++j)
        {
            ++anz;
            if (anz%8 == 0)
                fout1 << Sk[j] << endl;
            else if (anz%8 == 1)
                fout2 << Sk[j] << endl;
            else if (anz%8 == 2)
                fout3 << Sk[j] << endl;
            else if (anz%8 == 3)
                fout4 << Sk[j] << endl;
            else if (anz%8 == 4)
                fout5 << Sk[j] << endl;
            else if (anz%8 == 5)
                fout6 << Sk[j] << endl;
            else if (anz%8 == 6)
                fout7 << Sk[j] << endl;
            else if (anz%8 == 7)
                fout8 << Sk[j] << endl;
        }

        X = emptyset;
    }

    while((fin >> in).good())
    {
        X.add_vertex_lazy(in);
        ++counter;

        if (counter == startsetsize0)
        {
            X.update();

            Sk = augment_vertexset(X);

            for (int j=0; j<Sk.size(); ++j)
            {
                ++anz;
                if (anz%8 == 0)
                    fout1 << Sk[j] << endl;
                else if (anz%8 == 1)
                    fout2 << Sk[j] << endl;
                else if (anz%8 == 2)
                    fout3 << Sk[j] << endl;
                else if (anz%8 == 3)
                    fout4 << Sk[j] << endl;
                else if (anz%8 == 4)
                    fout5 << Sk[j] << endl;
                else if (anz%8 == 5)
                    fout6 << Sk[j] << endl;
                else if (anz%8 == 6)
                    fout7 << Sk[j] << endl;
                else if (anz%8 == 7)
                    fout8 << Sk[j] << endl;
            }

            X = emptyset;
            counter=0;
        }
    }
    if (!fin.eof())
    {
        cerr << "Error while reading file. Aborted." << endl;
        return 1;
    }
    if (X.setsize() != 0)
    {
        cerr << "Warning! Input file may be uncomplete!" << endl;
    }

    fin.close();

    fout1.close();
    fout2.close();
    fout3.close();
    fout4.close();
    fout5.close();
    fout6.close();
    fout7.close();
    fout8.close();


    graph_free(graph_symm_matrices);
    DYNFREE(g,g_sz);
    return 0;
}
