
#include "orderly.h"


using namespace std;

Vertexset::Vertexset(): sz(0), n(0)
{
    lab = (int*) 0;
    ptn = (int*) 0;
    orbits = (int*) 0;
    g_dist1 = (graph*) 0;
    g_min_dist = (graph_t *) 0;
}

Vertexset::Vertexset(graph* const graph_dist1, graph_t* graph_min_dist): sz(0), n(graph_min_dist->n), g_dist1(graph_dist1), g_min_dist(graph_min_dist)  // empty set
{
    int m = SETWORDSNEEDED(n);

    lab = new int [n];
    ptn = new int [n];
    for (int i=0; i<n; ++i)
    {
        lab[i]=i;
        ptn[i]=1; // cells end where the zeros are
    }

     // allocate memory to store the orbits of the automorphism group
    orbits = new int [n];

    this->update();
}


Vertexset::Vertexset(const Vertexset &X): sz(X.sz), n(X.n), g_dist1(X.g_dist1), g_min_dist(X.g_min_dist)
{
    lab = new int [n];
    ptn = new int [n];
    orbits = new int [n];
    for (int i=0; i<n; ++i)
    {
        lab[i] = X.lab[i];
        ptn[i] = X.ptn[i];
        orbits[i] = X.orbits[i];
    }
}


Vertexset Vertexset::operator=(const Vertexset &X)
{
    int m = SETWORDSNEEDED(X.n);

    if (n != X.n)
    {
        delete[] lab;
        delete[] ptn;
        delete[] orbits;

        n = X.n;

        lab = new int[n];
        ptn = new int[n];
        orbits = new int[n];
    }
    sz = X.sz;

    for (int i=0; i<n; ++i)
    {
        lab[i] = X.lab[i];
        ptn[i] = X.ptn[i];
        orbits[i] = X.orbits[i];
    }

    g_dist1 = X.g_dist1;
    g_min_dist = X.g_min_dist;

    return *this;
}

Vertexset Vertexset::add_vertex(int vertex)
{
    this->add_vertex_lazy(vertex);
    this->update();
    return *this;
}

Vertexset Vertexset::add_vertex_lazy(int vertex)
{
    int i=sz;
    while (lab[i] != vertex && i<n)
        ++i;

    if (i!=n)
    {
        lab[i] = lab[sz];
        lab[sz] = vertex;
        if (sz != 0)
            ptn[sz-1] = 1;
        ptn[sz] = 0;

        sz++;
    }
    return *this;
}

Vertexset Vertexset::update()
{
    int m = SETWORDSNEEDED(n);

    DEFAULTOPTIONS_GRAPH(options); //set options
    options.getcanon = TRUE; // canonical labelling
    options.defaultptn = FALSE; // use partitioning

    statsblk stats_add_vertex; // variable for densenauty statistics

    DYNALLSTAT(graph,g_can,g_can_sz);
    DYNALLOC2(graph,g_can,g_can_sz,m,n,"malloc");

    densenauty(g_dist1, lab, ptn, orbits, &options, &stats_add_vertex, m, n, g_can);
    DYNFREE(g_can, g_can_sz);

    return *this;
}

int Vertexset::setsize() const
{
    return sz;
}

int Vertexset::graph_size() const
{
    return n;
}

int Vertexset::get_lab(int i) const
{
    return lab[i];
}

int Vertexset::get_orbits(int i) const
{
    return orbits[i];
}

bool Vertexset::is_isom_to(const Vertexset &X) const
{
    int m = SETWORDSNEEDED(n);

    DYNALLSTAT(graph,g_can,g_can_sz);
    DYNALLOC2(graph,g_can,g_can_sz,m,n,"malloc");
    DYNALLSTAT(graph,g_canX,g_canX_sz);
    DYNALLOC2(graph,g_canX,g_canX_sz,m,n,"malloc");

    DEFAULTOPTIONS_GRAPH(options); //set options
    options.getcanon = TRUE; // canonical labelling
    options.defaultptn = FALSE; // use partitioning

    statsblk stats; // variable for densenauty statistics

    densenauty(g_dist1, lab, ptn, orbits, &options, &stats, m, n, g_can);
    densenauty(X.g_dist1, X.lab, X.ptn, X.orbits, &options, &stats, m, n, g_canX);


    for (int k = 0; k < m*(size_t)n; ++k)
        if (g_can[k] != g_canX[k])
        {
            DYNFREE(g_can, g_can_sz);
            DYNFREE(g_canX, g_canX_sz);
            return false;
        }

    DYNFREE(g_can, g_can_sz);
    DYNFREE(g_canX, g_canX_sz);
    return true;
}

double Vertexset::write_automs(string filename, bool writemarkers, bool append)
{
    int m = SETWORDSNEEDED(n);

    DEFAULTOPTIONS_GRAPH(options); //set options
    options.defaultptn = FALSE; // use partitioning
    options.writeautoms = TRUE; // write automorphisms
    options.writemarkers = writemarkers; // do (not) write additional information about automorphisms
    options.cartesian = TRUE;//FALSE; // cartesian representation of the automorphisms

    string mode;
    if (append)
        mode = "a";
    else
        mode = "w";
    FILE * file = fopen(filename.c_str(), mode.c_str());
    options.outfile = file;

    statsblk stats_autom; // variable for densenauty statistics

    densenauty(g_dist1, lab, ptn, orbits, &options, &stats_autom, m, n, NULL);

    fprintf(file, "\n group size: ");
    writegroupsize(file,stats_autom.grpsize1,stats_autom.grpsize2);

    fprintf(file, "\n orbits:\n");
    putorbits(file, orbits, 78, n);

    fclose (file);

    return stats_autom.grpsize1 * pow(10.0,stats_autom.grpsize2);
}

Vertexset::~Vertexset()
{
    delete[] orbits;
    delete[] ptn;
    delete[] lab;
}

ostream & operator<< (ostream & os, const Vertexset & X)
{
   // os << "{ ";
    for (int i=0; i<X.sz-1; ++i)
        os << X.lab[i] << " ";
    if (X.sz != 0)
        os << X.lab[X.sz-1];
 //   os << " }";

    return os;
}


vector<Vertexset> augment_vertexset(Vertexset X)
{
    vector<Vertexset> augmented_isom_classes;
    Vertexset Xux;

    int i,j,k,m,n,o;
    k = X.setsize();
    n = X.graph_size();
    m = SETWORDSNEEDED(n);
    for (j=k; j<n; ++j) {i = X.get_lab(j); if(i==X.get_orbits(i)) // for each orbit representative of G_X on V\X
    {
        bool adjacent = true;
        for (o=0; o<k; ++o)
            adjacent = adjacent && GRAPH_IS_EDGE(X.g_min_dist, i, X.get_lab(o));
        if (adjacent)
        {
            Xux = X;
            Xux.add_vertex(i);

            if(Xux.get_orbits(i)==Xux.get_orbits(Xux.get_lab(0)))
                augmented_isom_classes.push_back(Xux);
        }
    }}

    return augmented_isom_classes;
}
